#
# BSD 3-Clause License
#
# Copyright (c) 2022, California Institute of Technology and
# Max Planck Institute for Gravitational Physics (Albert Einstein Institute)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Tests for `from_string` method in `TDICombination` class.

Authors:
    Olaf Hartwig <olaf.hartwig@aei.mpg.de>
"""

from pytdi import TDICombination

tdiX1 = {
    "eta_21": [(-1, ["D_12"]), (1, ["D_13", "D_31", "D_12"])],
    "eta_12": [(-1, []), (1, ["D_13", "D_31"])],
    "eta_31": [(-1, ["D_12", "D_21", "D_13"]), (1, ["D_13"])],
    "eta_13": [(-1, ["D_12", "D_21"]), (1, [])],
}

tdiX2 = {
    "eta_21": [
        (-1, ["D_13", "D_31", "D_12"]),
        (-1, ["D_13", "D_31", "D_12", "D_21", "D_12"]),
        (1, ["D_12"]),
        (1, ["D_12", "D_21", "D_13", "D_31", "D_13", "D_31", "D_12"]),
    ],
    "eta_12": [
        (-1, ["D_13", "D_31"]),
        (-1, ["D_13", "D_31", "D_12", "D_21"]),
        (1, []),
        (1, ["D_12", "D_21", "D_13", "D_31", "D_13", "D_31"]),
    ],
    "eta_31": [
        (-1, ["D_13"]),
        (-1, ["D_13", "D_31", "D_12", "D_21", "D_12", "D_21", "D_13"]),
        (1, ["D_12", "D_21", "D_13"]),
        (1, ["D_12", "D_21", "D_13", "D_31", "D_13"]),
    ],
    "eta_13": [
        (-1, []),
        (-1, ["D_13", "D_31", "D_12", "D_21", "D_12", "D_21"]),
        (1, ["D_12", "D_21"]),
        (1, ["D_12", "D_21", "D_13", "D_31"]),
    ],
}


def test_empty():
    """Test if empty string returns empty combination"""
    combination = TDICombination.from_string("")
    assert combination.components == {}


def test_tdi_x1():
    """Check we build TDI X1 correctly."""
    combination = TDICombination.from_string("12131 -12131")
    assert combination.components == tdiX1


def test_tdi_x1_decomposed():
    """Check we build TDI X1 correctly with a decomposed string."""
    combination = TDICombination.from_string("12 21 13 31 -12 -21 -13 -31")
    assert combination.components == tdiX1


def test_tdi_x2():
    """Check we build TDI X2 correctly."""
    combination = TDICombination.from_string("121313121 -131212131")
    assert combination.components == tdiX2


def test_tdi_x2_decomposed():
    """Check we build TDI X2 correctly with a decomposed string."""
    combination = TDICombination.from_string("12 21 13 31 13 31 12 21 -131212131")
    assert combination.components == tdiX2
